// Distributed under GNU General Public License (see license.txt for details).
//
//  Copyright (c) 2013 Tianbao Yang.
//  All Rights Reserved.
//
// eagles fly on the cloud
// implements DisDCA for multi-class classification 
//
//*****************************************************************************
// Included Files
//*****************************************************************************
#include "cmd_line.h"
#include "data_set.h"
#include "objective.h"
#include "inc_dual.h"
#include <algorithm>
#include <functional>
#include <sstream>
#include <deque>
#include <set>
#include <iomanip>

//#include <mpi.h>
//using namespace MPI;



int main(int argc, char** argv) {


  // -------------------------------------------------------------
  // ---------------------- Parse Command Line -------------------
  // -------------------------------------------------------------

  std::string data_filename;
  std::string test_filename;
  std::string model_filename;
  std::string obj_filename;
  std::string label_filename;
  std::string test_label_filename;
  std::string class_filename;

  std::string sr;
  std::string ur;
  std::string format;
  std::string dtype;

  uint iter = 100;
  uint iiter=1;
  uint exam_per_iter = 1;
  uint iter_to_eval=0;
  uint dimension=0;
  uint batch=0;
  uint block=0;
  uint group=1;
  uint tebn=1;
  uint wsize=0;  
  
  int verb=0;
  int normalize  = 0;
  int ls =2;
  int org=1;

  double eta_constant=1;
  double l1s=0;
  double tol=0;
  double lambda = 1.0;

  // parse command line
  learning::cmd_line cmdline;
  cmdline.info("eagles for big data analytics (one vs all multi-class)");
  cmdline.add_master_option("<data-file or data-folder>", &data_filename);
  cmdline.add("-lambda", "regularization parameter (default = 1/n)", &lambda, 0);
  cmdline.add("-T", "number of iterations for optimizing SVM (default = 100)", &iter,100);
  cmdline.add("-ws", "window size (last ws number of iterations) for averaging (default = 0) (avg may be better, hint: -ws 200)", &wsize, 0);
  cmdline.add("-m", "size of block for stochastic gradient (default = 1000)", &exam_per_iter, 1000);
  cmdline.add("-verb", "display objective value or not (default = 0)", &verb, 0);
  cmdline.add("-tol", "tolerence parameter for inner iteration (default = None)", &tol, 0);
  cmdline.add("-norm", "normalize each data  or not (default = 1)", &normalize, 0);
  cmdline.add("-eval", "number of iterations to evalute objective(set to 0 or number of iterations to save time) (default = 0)", &iter_to_eval, 0);
  cmdline.add("-iiter", "number of iterations for optimizing sub-SVM (default = 1)", &iiter,1);
  cmdline.add("-eta", "the scaling factor for the learning rate (default = 1)", &eta_constant, 1);
  //cmdline.add("-sr", "the rule for computing the scalar s (R1 or R2) (default = R1)", &sr, "R1");
  //cmdline.add("-ur", "the update rule (bat or inc) (default = inc)", &ur, "inc");
  cmdline.add("-ls", "the loss function (1 (L1 hinge loss) or 2 (L2 hinge loss), 3 (logistic loss)) (default = 2)", &ls, 2);
  cmdline.add("-l1", "the scalar before l1 norm ", &l1s, 0);
  cmdline.add("-fmt","data format (dense or sparse(libsvm) default = sparse)", &format, "sparse");
  cmdline.add("-dtype","data type (txt or bin  default = NONE)", &dtype, "txt");
  cmdline.add("-nb","number of bacthes of data (default = number of procs)", &batch, 0);
  //cmdline.add("-nb","number of blocks of data (default = number of procs)", &block, 0);
  cmdline.add("-ng","number of groups of  of procs (default = 1)", &group, 1);
  cmdline.add("-nt","number of bacthes of test data (default = 1)", &tebn, 1);
  cmdline.add("-d","number of dimensions (for binary data option only)  (default = 0)", &dimension, 0);
  cmdline.add("-org","starting index of data (0, 1)  (default = 1)", &org, 1);
  cmdline.add("-labelFile","name of label foldername (for binary data option only,  assume files 1.txt, 2.txt) (default = noLabelFile)", &label_filename,"noLabelFile");
  cmdline.add("-testlabelFile","name of test label foldername (for binary data option only, assume files are 1.txt, 2.txt) (default = noLabelFile)", &test_label_filename,"noLabelFile");
  cmdline.add("-testFile","name of test data file foldername (default = noTestFile, assume files are 1.dat, 2.dat)", &test_filename,"noTestFile");
  cmdline.add("-classFile","name of target class file (default = noClassFile)", &class_filename,"noClassFile");
  cmdline.add("-objFile","name of objective file (default = noObjFile)", &obj_filename, "noObjFile");
  cmdline.add("-modelFile","name of model file (default = noModelFile)", &model_filename,"noModelFile");

  
  boost::mpi::environment env(argc, argv);
  boost::mpi::communicator WORLD;
  uint rank=WORLD.rank();
  uint size=WORLD.size();
  

  int rc = cmdline.parse(argc, argv);
  if (rc < 1) {
   if (rank==0)
   {
      cmdline.print_help();
      std::cout<<"number of argc="<<rc<<std::endl;
   }
   //Finalize();
   return EXIT_FAILURE;
  }


  /*if(ur!="inc" and ur!="bat"){
  	  std::cerr<<"ur = inc or bat"<<std::endl;
  	    return(EXIT_FAILURE);
    }
  if(sr!="R1" and ur!="R2"){
  	  std::cerr<<"sr = R1 or R2"<<std::endl;
  	  return(EXIT_FAILURE);
    }*/
  if(dtype=="dat" and dimension==0){
     std::cerr<<"dimension not given for binary data"<<std::endl;
     return(EXIT_FAILURE);
   }
  if(dtype=="dat" and label_filename =="noLabelFile"){
	     std::cerr<<"label file/folder name not given."<<std::endl;
	     return(EXIT_FAILURE);
  }
	  
  if(dtype=="dat" and test_filename!="noTestFile" and test_label_filename=="noLabelFile"){
     std::cerr<<"test label file/folder name not given. "<<std::endl;
     return(EXIT_FAILURE);

  }

  if(ls>3){
	     std::cerr<<"eagles only support loss: L1svm(ls=1), L2svm(l2=2), logistic Regression(ls=3)"<<std::endl;
	     return(EXIT_FAILURE);
  }
  
  std::string loss_type;
  if(ls==1) loss_type="SVM(hinge)";
  else if(ls==2) loss_type="SVM(squared-hinge)";
  else if(ls==3) loss_type="LogReg";
  else if(ls==4) loss_type="LeastReg";

  //--------------set the leader node-----------------------------
  if(size % group != 0){
    if(rank==0) std::cerr<<"#of groups of processors (ng) not divisible by #of processors (np)"<<std::endl;
    return(EXIT_FAILURE);
   }
  block = size/group;
  if(batch!=0 and block>batch) block=batch;

  uint my_group=0;
  std::vector<uint> class_leader_ranks;
  std::vector<uint> my_group_ranks;
  if(block>0 and block<size){
	group=floor((double)size/(double)block);
	my_group=floor((double)rank/(double)block);
  }else block=size;

  for(uint g=0; g<group; g++)
     class_leader_ranks.push_back(g*block);

  
  for(uint r=0; r<size; r++){
	  uint g=floor((double)r/(double)block);
	  if(g==my_group)
		  my_group_ranks.push_back(r);
  }

  boost::mpi::group mpi_group=WORLD.group(); 
  boost::mpi::group mpi_my_group=mpi_group.include(my_group_ranks.begin(), my_group_ranks.end());
  boost::mpi::group mpi_leader_group=mpi_group.include(class_leader_ranks.begin(), class_leader_ranks.end());
  boost::mpi::communicator COMM_WORLD(WORLD, mpi_my_group);



  uint my_rank=COMM_WORLD.rank();
  uint my_size=COMM_WORLD.size();
  uint my_class_leader=0;
  
  if(my_group>=group){
	  std::cout<<"processor: rank "<<rank<<" not in any group, exit"<<std::endl;
	  return(EXIT_SUCCESS);
  }
  
  if(my_rank==my_class_leader){
    if((uint)rank!=class_leader_ranks[my_group])
      std::cout<<"gloabl leader rank: "<<rank<<" != local leader rank: "<<class_leader_ranks[my_group]<<std::endl;
  }

  
  // -------------------------------------------------------------
  // ---------------------- read the data ------------------------
  // -------------------------------------------------------------
   std::string data_filename_r = data_filename;
   std::string label_filename_r = label_filename;
   if(size>1){
     std::ostringstream convert;
     convert<<my_rank+org;
     data_filename_r = data_filename + "/" + convert.str();
     if(dtype!="") data_filename_r =data_filename_r + "."+dtype;
     if(label_filename!="noLabelFile") label_filename_r = label_filename + "/" + convert.str() +".txt";
   }
   
   std::string task=(ls<=3)?"class":"regre";

   trainDataset Data(format, task, dimension, normalize==1); //(data_filename_r, format, task, dimension, normalize); 
   
   
   //if(batch!=0 and block>batch) block=batch;
   //if(block==0 or block>size) block=size;
   if(batch==0) {
      batch=block;
      if(rank==0)std::cerr<<"warning: check if #batches of data is "<<batch<<std::endl;
    }
   if(block>batch) {
    if(rank==0) std::cerr<<"# of procs per group > #of batches; exit "<<std::endl; 
    return(EXIT_FAILURE);
   }
   if(rank==0) std::cout<<"reading data..."<<std::flush;
   if(block==batch){
        Data.Append(data_filename_r, label_filename_r);
   }else{
       uint num_per_block=floor((double)batch/(double)block);
       uint start_batch=my_rank*num_per_block, end_batch=(my_rank+1)*num_per_block;
       if(my_rank+1==block and end_batch<batch) end_batch=batch;
       for(uint b=start_batch; b<end_batch and b<batch; b++){
          std::ostringstream convert;
          convert<<b+org;
          data_filename_r = data_filename + "/" + convert.str();
          if(dtype!="") data_filename_r =data_filename_r + "."+dtype;
          if(label_filename!="noLabelFile")  label_filename_r=label_filename +"/"+convert.str()+".txt"; 
          Data.Append(data_filename_r, label_filename_r); 
        }
   }

     ulong ns=Data.size();
     uint ds=Data.dim();
     double Rs=Data.mR();
     long readingTime=Data.readtime();

     
     
     ulong n;
     double R; 
     all_reduce(COMM_WORLD, ns, n, std::plus<ulong>());
     all_reduce(COMM_WORLD, Rs, R, boost::mpi::maximum<double>());
     all_reduce(COMM_WORLD, ds, dimension, boost::mpi::maximum<uint>());

     if(rank==0 and (R==0 || ns==0 || n==0)){
       std::cerr<<"something wrong in the data"<<std::endl;
       std::cerr<<"max square norm="<<R<<",ns="<<ns<<",n="<<n<<std::endl;
       return(EXIT_FAILURE);
     }  
     
       Dataset testData(format, task, dimension, normalize==1); //(data_filename_r, format, task, dimension, normalize); 
       long testReadingTime=0;
       if (test_filename != "noTestFile" && rank==0) {
     	  if(tebn==1){
            testData.Append(test_filename, test_label_filename);
     	  }else{
     		 for(uint b=0; b<tebn; b++){
     		      std::string test_filename_r, test_label_filename_r;
     		      std::ostringstream convert;
     		      convert<<b+org;
     		      test_filename_r = test_filename + "/" + convert.str();
     		      if(dtype!="") test_filename_r=test_filename_r + "."+dtype;
     		      if(test_label_filename!="noLabelFile")  test_label_filename_r=test_label_filename +"/"+convert.str()+".txt"; 
     		      testData.Append(test_filename_r, test_label_filename_r);
     		 }
     	  }
     	testReadingTime=testData.readtime();  
      }else{
         testReadingTime = 0;
       }
       long my_readingTime = readingTime + testReadingTime;
       
       if(rank==0){
         std::cout<< my_readingTime<< "s" <<std::endl<<std::flush;
       }  
    
     std::set<int> uLabels;
     if(task=="class")  uLabels=Data.mclass();
     if(uLabels.size()==2){
       if(rank==0) std::cerr<<"#of classes = 2, big_birds is large,  consider eagles\n"; 
     }


  
     //----------syn class labels---------------------------
          std::vector<int> my_Class;
          std::vector<int> Class;
          uint nc, tc; 
          if(class_filename!="noClassFile"){
          	std::ifstream cla(class_filename.c_str()); 
          	if (!cla.good()) {
          	    std::cerr << "error w/ " << class_filename << std::endl;
          	    exit(EXIT_FAILURE);
          	}
              std::string line;
              while(std::getline(cla, line))
              {
                std::istringstream in(line);
                int label; 
                in>>label;
                Class.push_back(label);
              }
          	tc=Class.size();
          	cla.close();
          }else{
             std::vector<int>ulabels(uLabels.begin(), uLabels.end());
          
             uint max_c;
             all_reduce(COMM_WORLD, (uint) ulabels.size(), max_c, boost::mpi::maximum<uint>());
             ulabels.resize(max_c, ulabels[0]);
             Class.resize(max_c*my_size,0);
             gather(COMM_WORLD, ulabels.data(), max_c, Class, 0);
             if(my_rank==0){
             std::sort(Class.begin(), Class.end()); 
             auto last = std::unique(Class.begin(), Class.end());
             Class.erase(last, Class.end());
             tc=Class.size();
             }
             broadcast(COMM_WORLD, tc, 0);
             Class.resize(tc, 0);
             broadcast(COMM_WORLD, Class, 0);   
          }
          
             if(group==1)
          	   my_Class=Class;
             else{
          	   int nc_per_group=floor((double)tc/(double)group);
          	   if(nc_per_group==0) nc_per_group=1; 	   
          	   std::vector<int> group_cn(group, nc_per_group); // group class number
          	   if(nc_per_group*group<tc){
          		 uint rc=tc-nc_per_group*group;
          		   for(uint g=0; g<group; g++){
          			  if((g+1)<=rc) group_cn[g] +=1;
          		   }
          	   }
          	   uint start_i, end_i;
          	   if(my_group==0) start_i=0;
          	   else start_i = std::accumulate(group_cn.data(), group_cn.data()+my_group, 0) ;
          	   end_i=start_i + group_cn[my_group];


          	   for(uint i=start_i; i<end_i and i<tc; i++)
          	   	  my_Class.push_back(Class[i]);   
             }
             nc=my_Class.size();
             if(nc==0){
               std::cout<<"processor: rank="<<rank<<", my group="<<my_group+1<<", #of classes="<<tc<<", no class, exit"<<std::endl;
               return(EXIT_FAILURE);
              }
             if(group>tc)
             {
               for(uint g=tc; g<group; g++)
                 class_leader_ranks.pop_back();
             }
           /*if(rank==0){
            std::cout<<"rank ="<<rank<<", my group="<<my_group+1<<", #of classes="<<nc<<std::endl;
            std::copy(my_Class.begin(), my_Class.end(), std::ostream_iterator<int>(std::cout, "|"));
            std::cout<<std::endl;
           }*/
             
          if(verb and rank==0){ 
           std::cout << n << " = number of training data\n"<<std::flush;
           std::cout << tc << " = number of classes\n"<<std::flush;
          }
        
    

  // choose a random seed
  srand(1);
  // ---------------start loop here--------------------------------
  // --------------------------------------------------------------
  // -------------------------------------------------------------
  // -------------------------------------------------------------
  long train_calc_obj_time[2]={0,0}, iter_calc_obj_time, total_train_calc_obj_time[2]={0,0};

  //to output the objective using both w and averaged w
  std::vector<double> phi_loss_norm(5*nc, 0);
  std::vector<double>total_phi_loss(3*nc,0);
  
  std::vector<double> test_error(1+(wsize>0?1:0), 0);
  
  //----------initialize-------------------------------------------
  std::vector<WeightVector> Wc, Wrc, AvgWc;
  std::vector<std::vector<double> > Alphac, obj_vecc, dual_gap_vecc;
  std::vector<double> obj_vec, dual_gap_vec;
  dual_gap_vec.reserve(nc);
  for(uint c=0; c<nc; c++)
  {
	  WeightVector W(dimension), Wr(dimension), AvgW(dimension);
	  std::vector<double> Alpha(ns, 0.0);
	  Wc.push_back(W);
	  Wrc.push_back(Wr);
	  AvgWc.push_back(AvgW);
	  Alphac.push_back(Alpha);
	  
	  if(ls<=2){
	    obj_vec.push_back(1.0); if(wsize)  obj_vec.push_back(1.0);
	    dual_gap_vec.push_back(1.0);
	  }
	  else if(ls==3){
		obj_vec.push_back(log(2)); if(wsize) obj_vec.push_back(log(2)); 
		dual_gap_vec.push_back(log(2));
	  }
  }
  obj_vecc.push_back(obj_vec);
  dual_gap_vecc.push_back(dual_gap_vec);


  if(lambda==0) lambda=1.0/n;
  if(exam_per_iter==0 or exam_per_iter>ns) exam_per_iter=ns;
  if(tol>0 and iter_to_eval==0) iter_to_eval = iter; // every iter to evaluate the objective

  double coeff=1.0*my_size/(n*lambda);
  double conj_val=0;
  //double beta = exam_per_iter;

  std::vector<ulong> index;
  for(ulong i=0; i<ns; i++) index.push_back(i); // handle the difference size of different portions of data
  ulong shift=0; // use to index the element for updating
  uint count=0; // count for counting the number of averaged W

 
  double max_dual_gap;
  //---------------Main Loop---------------------------------
  long startTime = get_runtime(), endTime;
  uint i=0, iter_incre=iter;
  if(verb and rank==0) std::cout<<"optimize model..."<<std::endl<<std::flush; 
  while(i<iter){
      dual_update(ls, Data, my_Class, Wc, Wrc, Alphac, coeff,  phi_loss_norm, iiter, exam_per_iter, index, shift);
      all_reduce(COMM_WORLD, Wrc, Wc, vecW_plus());
	  for(uint c=0; c<nc; c++) {
		   //--non serialized version
		   //double *wr=Wrc[c].make_my_a_one();
		   //double *w=Wc[c].make_my_a_one();
		   //all_reduce(COMM_WORLD, wr, dimension, w, std::plus<double>());	 
	       Wc[c].scale(1.0/my_size);
	       if(l1s>0.0) conj_val = Wc[c].proximal_l1(l1s);
	       if(wsize and iter-i<=wsize){
	        AvgWc[c].add(Wc[c],1.0);
	     }
	   }
	   if(wsize and iter-i<=wsize) count+=1;

	   //calcuate obj
       if(iter_to_eval && (i+1) % iter_to_eval==0 ){
	   long iter_calc_start_time=get_runtime();
       objective(Data, my_Class, iter_calc_obj_time, phi_loss_norm, Wc, AvgWc, count, count, ls, l1s);

       int reduce_size=2*nc; if(wsize>0 and count>0) reduce_size=3*nc;
       all_reduce(COMM_WORLD, phi_loss_norm.data(), reduce_size, total_phi_loss.data(), std::plus<double>());
       max_dual_gap=0;
       compute_obj_dual_gap(n, nc, phi_loss_norm, total_phi_loss, conj_val, lambda, l1s, wsize, count,  obj_vecc, dual_gap_vecc, max_dual_gap);
  	   long iter_calc_end_time=get_runtime();
	   iter_calc_obj_time = iter_calc_end_time - iter_calc_start_time;     
       }
       else iter_calc_obj_time=0;
       
      endTime=get_runtime();
      train_calc_obj_time[0] +=endTime - startTime - iter_calc_obj_time;
      train_calc_obj_time[1] +=iter_calc_obj_time;
      startTime = get_runtime();

      i++;
      if(i==iter){
	   if(tol>0 and max_dual_gap>tol){
	       iter +=iter_incre;
           if(verb and rank==0 and i%iter_to_eval==0) std::cout<<"max dual gap="<<max_dual_gap<<std::endl;
	   }else if(max_dual_gap<tol)
       {iter=i; break;}
	   else if(tol==0) break;
      }
}// end for(int i=0; ...)


 if(verb and rank==0) std::cout<<"optimize model done!"<<std::endl<<std::flush; 
 if(wsize) 
	 for(uint c=0; c<nc; c++)
		 AvgWc[c].scale(1.0/count);
 
 endTime=get_runtime();
 train_calc_obj_time[0] += endTime - startTime;

 //COMM_WORLD.Allreduce(train_calc_obj_time, total_train_calc_obj_time, 2, LONG, SUM);
 all_reduce(COMM_WORLD, train_calc_obj_time, 2, total_train_calc_obj_time, std::plus<long>());


 if(iter_to_eval ==0 or i<iter_to_eval){
	  if(verb and rank==0) std::cout<<"comp obj...!"<<std::endl<<std::flush; 
      objective(Data, my_Class, iter_calc_obj_time, phi_loss_norm, Wc, AvgWc, count, count, ls, l1s);

	  int reduce_size=2*nc; if(wsize>0 and count>0) reduce_size=3*nc;
          all_reduce(COMM_WORLD, phi_loss_norm.data(), reduce_size, total_phi_loss.data(), std::plus<double>());
	      max_dual_gap=0;
          compute_obj_dual_gap(n, nc, phi_loss_norm, total_phi_loss, conj_val, lambda, l1s, wsize, count, obj_vecc, dual_gap_vecc, max_dual_gap);
          if(verb and rank==0) std::cout<<"comp obj done!"<<std::endl<<std::flush; 


 }

 if(obj_filename != "noObjFile" && my_rank==my_class_leader){
  if(verb and rank==0) std::cout<<"print obj..."<<std::endl<<std::flush; 
  std::string obj_filename_c;
  std::ostringstream convert;
  convert<<my_Class.front();
  obj_filename_c = obj_filename + "." + convert.str() ;
  std::ofstream obj(obj_filename_c.c_str());
  for(uint it=0; it<obj_vecc.size(); it++){
    copy(obj_vecc[it].begin(), obj_vecc[it].end(), std::ostream_iterator<double>(obj, " "));
    obj<<std::endl;
  }
  obj.close();
  
  std::string dual_filename_c;
  dual_filename_c =  "dual." + convert.str() ;
  std::ofstream dual(dual_filename_c.c_str());
  for(uint it=0; it<dual_gap_vecc.size(); it++){
    copy(dual_gap_vecc[it].begin(), dual_gap_vecc[it].end(), std::ostream_iterator<double>(dual, " "));
    dual<<std::endl;
  }
  dual.close();

  if(verb and rank==0) std::cout<<"print obj done!"<<std::endl<<std::flush; 
 }
 
 
 if(verb and rank==0) std::cout<<"collect model..."<<std::endl<<std::flush; 
 if(my_rank==my_class_leader and rank!=0){
	if(wsize){     
	   boost::mpi::request reqs[2];
	   reqs[0]=WORLD.isend(0, rank, Wc);
	   reqs[1]=WORLD.isend(0, rank+class_leader_ranks.back(), AvgWc); 
       boost::mpi::wait_all(reqs, reqs + 2);
	 }
	 else WORLD.send(0, rank, Wc);
 }
 if(rank==0){
	 for(std::vector<uint>::iterator it=class_leader_ranks.begin(); it!=class_leader_ranks.end(); it++){
		 if(*it!=(uint)rank){
		   if(wsize){
              boost::mpi::request reqs[2];
		      std::vector<WeightVector> rWc;
		      reqs[0]=WORLD.irecv(*it, *it, rWc);
	          std::vector<WeightVector> rAvgWc;
	          reqs[1]=WORLD.irecv(*it, *it + class_leader_ranks.back(), rAvgWc);
	          boost::mpi::wait_all(reqs, reqs + 2);
		      Wc.insert(Wc.end(), rWc.begin(), rWc.end());
		      AvgWc.insert(AvgWc.end(), rAvgWc.begin(), rAvgWc.end());
		   }
		   else{
		      std::vector<WeightVector> rWc;
		      WORLD.recv(*it, *it, rWc);
		      Wc.insert(Wc.end(), rWc.begin(), rWc.end());
		   }
	      }
	 }
 }	 
 if(verb and rank==0) std::cout<<"collect model done!"<<std::endl<<std::flush; 

 
 if (test_filename != "noTestFile" and rank==0) {
// need to gather all W to evaluate 
       if(verb and rank==0) std::cout<<"evalute test..."<<std::endl<<std::flush; 
	   objective(testData, Class,  iter_calc_obj_time,  test_error, Wc, AvgWc, wsize, ls);
       if(verb and rank==0) std::cout<<"evalute test done!"<<std::endl<<std::flush; 
 }
 
 if (model_filename != "noModelFile" and rank==0) {
	   std::string class_filename_c;
	   class_filename_c =  "eagles_classes.txt" ;
	   std::ofstream cla(class_filename_c.c_str());
	   copy(Class.begin(), Class.end(), std::ostream_iterator<int>(cla, " "));
	   cla<<std::endl;
	   cla.close();	 
	
   if(verb and rank==0) std::cout<<"print model..."<<std::endl<<std::flush; 
   std::ofstream model_file(model_filename.c_str(), std::ofstream::ate);
   if (!model_file.good()) {
     std::cerr << "error w/ " << model_filename << std::endl;
     exit(EXIT_FAILURE);
   }
   for(uint c=0; c<tc; c++)
      Wc[c].print(model_file);
   if(wsize){
      model_file<<"#---------avg model--------"<<std::endl;
      for(uint c=0; c<tc; c++)
        AvgWc[c].print(model_file);
   }
   model_file.close();
   if(verb and rank==0) std::cout<<"print model done!"<<std::endl<<std::flush; 
   
 }
  // -------------------------------------------------------------
  // ---------------------- Print Results ------------------------
  // -------------------------------------------------------------
if(my_rank==my_class_leader){ 
      std::ostringstream out (std::ostringstream::ate);  
      out<<std::left<<"+--------------------------------------------+\n";
      out<<"running summary: "<<loss_type<<"\n";
      out<<std::setw(35)<<std::left<<"leader rank:"<<rank<<std::endl
               <<std::setw(35)<<std::left<<"#classes:"<<nc<<std::endl
               <<std::setw(35)<<std::left<<"classes:"<<"{"<<my_Class.front()<<"..."<<my_Class.back()<<"}"<<std::endl
               <<std::setw(35)<<std::left<<"#examples(n):"<<n<<std::endl
               <<std::setw(35)<<std::left<<"#features(d):"<<dimension<<std::endl
               <<std::setw(35)<<std::left<<"#examples-per-iteration(-m):"<<exam_per_iter<<std::endl
               <<std::setw(35)<<std::left<<"#processors(-np):"<<size<<std::endl
               <<std::setw(35)<<std::left<<"#iterations(-T):"<<iter<<std::endl
               <<std::setw(35)<<std::left<<"training-time(s):"<<total_train_calc_obj_time[0]/size<<std::endl
               <<std::setw(35)<<std::left<<"evaluation-time(s):"<<total_train_calc_obj_time[1]/size<<std::endl
               <<std::setw(35)<<std::left<<"regularization param(-lambda):"<<lambda<<std::endl
               <<std::setw(35)<<std::left<<"max duality gap:"<<max_dual_gap<<std::endl;
      if(test_filename!="noTestFile" and rank==0){
         out<<std::setw(35)<<std::left<<"test error: last solution:"<<test_error[0]/testData.size()<<std::endl;
         if(wsize)
          out<<std::setw(35)<<std::left<<"test error: avged solution:"<<test_error[1]/testData.size()<<std::endl;
       }

     /*std::ostringstream out (std::ostringstream::ate);  
     out.width(10);
     out.fill(' ');
     out<<std::left;
     out<<"+--------------------------------------------+\n";
     out<<"running summary: "<<loss_type<<"\n";
     if(nc>1){
          out<<"leader		#classes	classes\n"
             <<std::setw(16)<<rank<<std::setw(16)<<nc<<"["<<my_Class.front()<<"..."<<my_Class.back()<<"]\n";
     }
     else
	      out<<"leader		class		max_gap\n"<<std::setw(16)<<rank<<std::setw(16)<<my_Class.back()<<max_dual_gap<<std::endl;
     
    out<<"examples(n)	dimensions(d)	lambda\n"
           <<std::setw(16)<<n<<std::setw(16)<<dimension<<lambda<<std::endl;
    
    out<<"process(K)	exam_per(m)	iter_to_eval\n"
           <<std::setw(16)<<size<<std::setw(16)<<exam_per_iter<<iter_to_eval<<std::endl;

    out <<"Iters		Train-time	Eval-Time\n"
              <<std::setw(16)<<iter<<std::setw(16)<<total_train_calc_obj_time[0]/my_size
              <<total_train_calc_obj_time[1]/my_size<<std::endl;
   */
   /*if(nc<=4){
	        out <<"Norm of ws\n";
            std::copy(phi_loss_norm.data()+3*nc, phi_loss_norm.data()+4*nc, std::ostream_iterator<double>(out," "));
   } else {
            out <<"Norm of ws\n";
            std::copy(phi_loss_norm.data()+3*nc, phi_loss_norm.data()+std::min(4*nc,3*nc+2), std::ostream_iterator<double>(out," "));
   }

   if(wsize){
      if(nc<=4){
        	out <<"Norm of avg ws\n";
            std::copy(phi_loss_norm.data()+4*nc, phi_loss_norm.data()+5*nc, std::ostream_iterator<double>(out," "));
	    }
      else{
    	     out <<"Norm of avg ws\n";
            std::copy(phi_loss_norm.data()+4*nc, phi_loss_norm.data()+std::min(5*nc, 4*nc+2), std::ostream_iterator<double>(out," "));
      }
   }
   if(nc<=4){
            out<<"\nprimal of ws (and avg ws)\n";
            std::copy(obj_vecc.back().begin(), obj_vecc.back().end(), std::ostream_iterator<double>(out," "));
            out<<"\ngap of Ws\n";
            std::copy(dual_gap_vecc.back().begin(), dual_gap_vecc.back().end(), std::ostream_iterator<double>(out," "));
            out<<std::endl;
	    }  else  {
            out<<"\nprimal of ws (and/or avg ws)\n";
            std::copy(obj_vecc.back().data(), obj_vecc.back().data()+std::min(2, (int)obj_vecc.back().size()), std::ostream_iterator<double>(out," "));
            out<<"\ngap of Ws\n";
            std::copy(dual_gap_vecc.back().data(), dual_gap_vecc.back().data()+std::min(2, (int)dual_gap_vecc.back().size()), std::ostream_iterator<double>(out," "));
	        out<<"max gap\n"<<max_dual_gap<<std::endl;
	    }*/
  	    
   /*if(test_filename!="noTestFile" and rank==0){
	    if(nc>1){
	     if(wsize){ out<<"max_gap		te_error(w)		te_error(avgw)\n"
	    	        <<std::setw(16)<<max_dual_gap<<std::setw(16)<<test_error[0]/testData.size()
	    	        <<test_error[1]/testData.size()<<std::endl;
	     } else  out<<"max_gap		test_error	\n"<<std::setw(16)<<max_dual_gap<<std::setw(16)
	                <<test_error[0]/testData.size()<<std::endl;
	    } else {
		     if(wsize){ out<<"te_error(w)		te_error(avgw)\n"
		    	        <<std::setw(16)<<test_error[0]/testData.size()
		    	        <<test_error[1]/testData.size()<<std::endl;
		     } else  out<<"test_error	"<<std::setw(16)<<test_error[0]/testData.size()<<std::endl;	
	    }
    } else if( test_filename =="noTestFile" and nc>1){
    	out<<"max_gap		"<<max_dual_gap<<std::endl;
    }
     out<<"+--------------------------------------------+\n";
     std::cout<<out.str()<<std::flush<<std::endl;
   */
    out<<"+--------------------------------------------+\n";
    std::cout<<out.str()<<std::flush<<std::endl;
 }
 return(EXIT_SUCCESS);
}















