// Distributed under GNU General Public License (see license.txt for details).
//
//  Copyright (c) 2013 Tianbao Yang.
//  All Rights Reserved.
//==============================================================================
// File Name: dense_vec.h
// implement dense vector 
//==============================================================================

#ifndef _DENSE_VEC
#define _DENSE_VEC

//*****************************************************************************
// Included Files
//*****************************************************************************
#include <iostream>
#include <fstream>
#include <cstdio>
#include <cstdlib>
#include <sstream>

typedef unsigned int 	uint;


/** Implements a dense vector
*/
class dense_vector {
 public:

  /** Default Constructor. Allocates an all zeros sparse vector
  */
  dense_vector(): d(0), my_v(NULL)
  {  }

  /** Constructor. Construct a vector from an istringstream.
      @param is The input stringstream to read from.
      @param n The number of elements to read
  */
  dense_vector(uint dd): d(dd), my_snorm(0), my_v(NULL)
  {
	  //my_v=new double[d];
	  my_v=new float[d];
	  // read the elements
	  for (uint i=0; i<d ; ++i) {
        my_v[i]=0;
	  }
  }
  

  
  dense_vector(std::istringstream& is, uint dd, bool index_flag=false): d(dd), my_snorm(0), my_v(NULL)
  {
	  //my_v=new double[d];
	  my_v=new float[d];
	  // read the elements
	  for (uint i=0; i<d ; ++i) {
		//double val;
		float val;
		is >> val;
        my_v[i]=val;
        my_snorm +=val*val;
	  }
  }
  
  // a copy constructor needed storing the class object in vector array
  dense_vector(const dense_vector& v){
    d=v.d;
    //my_v =new double[d];
    my_v =new float[d];
    my_snorm=v.my_snorm;
    for(uint i=0; i<d; ++i)
 	     my_v[i]=v.my_v[i];
  }
  
  // destructor
  ~dense_vector() { delete[] my_v;  }

  
  template<typename T>
  void initialize(uint dd, std::vector<T>& vec)
  {
	  if((uint)vec.size()!=dd){
		  std::cerr<<"initialize dense_vec size not consistent!"<<std::endl;
		  exit(EXIT_FAILURE);
	  }
	  if(my_v!=NULL) delete[] my_v;
	  d=dd;
	  my_snorm=0;
	  //my_v=new double[d];
	  my_v=new float[d];
	  
	  // read the elements
	  for (uint i=0; i<d ; ++i) {
        my_v[i]=(float)vec[i];
        my_snorm +=my_v[i]*my_v[i];
	  }
  }
 
  /** Returns the squared l_2 norm of the vector
      @return the squared l_2 norm of the vector
  */
  double snorm()
  {
	  return my_snorm;
  }

  float operator[](uint i) const{
    if (i < d) return my_v[i];
    return 0.0;
  }
  uint dimension(){
	  return d;
  }
  
  // Multiply a dense vector by a scalar s
  void scale(double s){
	my_snorm *= (s*s); 
	for (uint i=0; i<d; ++i) my_v[i] *=s ;
  }
  
  // normalize a dense vector
  void normalize(){
	for (uint i=0; i<d; ++i) my_v[i] /=std::sqrt(my_snorm) ;
	my_snorm = 1; 
  }
  
  // Print the content of a sparse instance to an output stream
  void print(std::ostream& os)
  {
      //std::copy(my_v, my_v+d, std::ostream_iterator<double>(std::cout, " "));
      std::copy(my_v, my_v+d, std::ostream_iterator<float>(std::cout, " "));
      std::cout<<std::endl;
  }
  
 private:  
  // The internal representation of w is as w = a*v where:
  uint d;
  double my_snorm;
  //double* my_v;
  float* my_v; 

};


#endif
